
#include "gdkbuffergap.h"
#include <glib.h>
#include <string.h>

/* Static functions declarations: */

/* grow_gap()
Description: Increases the size of the gap by 
		sizeof(element type) * 'element_count' bytes.
Inputs: bufgap: The GdkBufferGap whose gap you which to grow.
	element_count: The number of elements to grow by
Outputs: A pointer to the new bufgap, or NULL upon error
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: The bufgap->buffer is realloced to have a bigger gap.
*/
static GdkBufferGap *grow_gap(GdkBufferGap *bufgap, guint element_count);

/* End static function declarations. */



#define DEFAULT_GAP_GROWTH_SIZE 1024

GdkBufferGap *gdk_buffer_gap_new(GdkBufferGapBufferType buffer_type) {

	GdkBufferGap *new_bufgap;	/* The returned value */
	guint byte_count;		/* The number of bytes to allocate in 
					   the buffer */

	new_bufgap = (GdkBufferGap *)g_malloc( sizeof(GdkBufferGap) );


	/* DATATYPE code: */
	switch ( buffer_type ) {
	
		case gucharBuffer:
			byte_count = sizeof(guchar) * DEFAULT_GAP_GROWTH_SIZE;
			new_bufgap->buffer.ch = (guchar *)g_malloc(byte_count);
			break;

		case GdkWCharBuffer:
			byte_count = sizeof(GdkWChar) * DEFAULT_GAP_GROWTH_SIZE;
			new_bufgap->buffer.wc = (GdkWChar *)g_malloc(byte_count);
			break;
			
		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return NULL;
	}
	/* End DATATYPE code */

	/* Set default values: */
	new_bufgap->buffer_size = DEFAULT_GAP_GROWTH_SIZE;

	new_bufgap->buffer_type = buffer_type;

	new_bufgap->gap_position = 0;

	new_bufgap->gap_size = DEFAULT_GAP_GROWTH_SIZE;

	new_bufgap->gap_growth_size = DEFAULT_GAP_GROWTH_SIZE;

	return new_bufgap;
}  /* End gdk_buffer_gap_new() */



void gdk_buffer_gap_free(GdkBufferGap *bufgap) {

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
	
		case gucharBuffer:
			g_free(bufgap->buffer.ch);
			break;

		case GdkWCharBuffer:
			g_free(bufgap->buffer.wc);
			break;
			
		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
	}
	/* End DATATYPE code */

	g_free(bufgap);

}  /* End gdk_buffer_gap_free() */



GdkBufferGap *gdk_buffer_gap_insert(GdkBufferGap *bufgap, 
					gconstpointer chars,
					guint num_chars ) {

	gpointer dest;	/* The mem address to copy the chars to */
	guint byte_count;	/* The number of bytes to copy */
	
	g_return_val_if_fail(bufgap != NULL, NULL);
	g_return_val_if_fail(chars != NULL, NULL);

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
		case gucharBuffer:
			dest = (gpointer)(bufgap->buffer.ch + bufgap->gap_position);	
			byte_count = sizeof(guchar) * num_chars;
			break;

		case GdkWCharBuffer:
			dest = (gpointer)(bufgap->buffer.wc + bufgap->gap_position);	
			byte_count = sizeof(GdkWChar) * num_chars;
			break;
			
		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return NULL;
	}
	/* End DATATYPE code */

	if( num_chars >= bufgap->gap_size ) {
		/* Grow the gap to hold the data */
		grow_gap(bufgap, 
			num_chars + bufgap->gap_growth_size - bufgap->gap_size);
	}	

	/* Now that we know the gap is big enough to hold it, insert the data */
	g_memmove(dest, chars, byte_count);
	bufgap->gap_position += num_chars;
	bufgap->gap_size -= num_chars;
	
	/* Is there a more useful return value? */
	return bufgap;
					
}  /* End gdk_buffer_gap_insert() */



guint gdk_buffer_gap_delete(GdkBufferGap *bufgap, 
				guint num_chars, 
				gboolean forward) {

	guint right_size;	/* Number of data chars on the right */
	guint left_size;	/* Number of data chars on the left */

	right_size = bufgap->buffer_size - (bufgap->gap_size + bufgap->gap_position);
	left_size = bufgap->gap_position;

	if (forward == TRUE) {
		/* Clip to the logical maximum of characters */
		num_chars = MIN(num_chars, right_size);
		bufgap->gap_size += num_chars;
	}
	else {  /* backward */
		/* Clip to the logical maximum of characters */
		num_chars = MIN(num_chars, left_size);
		bufgap->gap_size += num_chars;
		bufgap->gap_position -= num_chars;
	}

	return num_chars;

}  /* End gdk_buffer_gap_delete() */


guint gdk_buffer_gap_move_gap(GdkBufferGap *bufgap, guint position) {

	gpointer src;		/* Source of the memmove */
	gpointer dest;		/* Destination of the memmove */
	guint byte_count;	/* Number of bytes to move */

	guint delta_x;	/* number of spaces that we're moving, and hence,
			   the number of data elements that are memmoved */
	

	/* First, clip to the logical maximum position */
	position = MIN(position, gdk_buffer_gap_get_data_size(bufgap) );
	
	
	if (position > bufgap->gap_position) {
		/* Moving to the right...*/
		delta_x = position - bufgap->gap_position ;
	
		/* DATATYPE code: */
		switch ( bufgap->buffer_type ) {
			case gucharBuffer:
				dest = (gpointer)(bufgap->buffer.ch 
						  + bufgap->gap_position);
				src = (gpointer)(bufgap->buffer.ch
						 + bufgap->gap_position 
						 + bufgap->gap_size);	
				byte_count = sizeof(guchar) * delta_x;
				break;

			case GdkWCharBuffer:
				dest = (gpointer)(bufgap->buffer.wc 
						  + bufgap->gap_position);
				src = (gpointer)(bufgap->buffer.wc
						 + bufgap->gap_position 
						 + bufgap->gap_size);	
				byte_count = sizeof(GdkWChar) * delta_x;
				break;
			
			default:
				g_warning("Unrecognized GdkBufferGapBufferType! ");
				return position;
		}
		/* End DATATYPE code */

		g_memmove(dest, src, byte_count);
		bufgap->gap_position = position;
	} /* (end if) */

	/* "else" */

	if (bufgap->gap_position > position) {
		/* Moving to the left...*/
		delta_x = bufgap->gap_position - position;
	
		/* DATATYPE code: */
		switch ( bufgap->buffer_type ) {
			case gucharBuffer:
				dest = (gpointer)(bufgap->buffer.ch 
						  + position 
						  + bufgap->gap_size);
				src = (gpointer)(bufgap->buffer.ch
						 + position);
				byte_count = sizeof(guchar) * delta_x;
				break;

			case GdkWCharBuffer:
				dest = (gpointer)(bufgap->buffer.wc 
						  + position 
						  + bufgap->gap_size);
				src = (gpointer)(bufgap->buffer.wc
						 + position);
				byte_count = sizeof(GdkWChar) * delta_x;
				break;
			
			default:
				g_warning("Unrecognized GdkBufferGapBufferType! ");
				return position;
		}
		/* End DATATYPE code */

		g_memmove(dest, src, byte_count);
		bufgap->gap_position = position;
	} /* (end if) */
	
	return position;
					
}  /* End gdk_buffer_gap_move_gap() */


gconstpointer gdk_buffer_gap_get_left_pointer(GdkBufferGap *bufgap) {

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
		case gucharBuffer:
			return (gconstpointer)(bufgap->buffer.ch);

		case GdkWCharBuffer:
			return (gconstpointer)(bufgap->buffer.wc);

		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return NULL;
	}
	/* End DATATYPE code */

}  /* End gdk_buffer_gap_get_left_pointer() */


guint gdk_buffer_gap_get_left_size(GdkBufferGap *bufgap) {

	return (bufgap->gap_position);

}  /* End gdk_buffer_gap_get_left_size() */



gconstpointer gdk_buffer_gap_get_right_pointer(GdkBufferGap *bufgap) {

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
		case gucharBuffer:
			return (gconstpointer)(bufgap->buffer.ch
						+ bufgap->gap_position
						+ bufgap->gap_size);

		case GdkWCharBuffer:
			return (gconstpointer)(bufgap->buffer.wc
						+ bufgap->gap_position
						+ bufgap->gap_size);

		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return NULL;
	}
	/* End DATATYPE code */

}  /* End gdk_buffer_gap_get_right_pointer() */



guint gdk_buffer_gap_get_right_size(GdkBufferGap *bufgap) {

	return (bufgap->buffer_size - bufgap->gap_position - bufgap->gap_size);

}  /* End gdk_buffer_gap_get_right_size() */


GdkBufferGap *gdk_buffer_gap_set_gap_growth_size(GdkBufferGap *bufgap, 
				guint element_count) {
	bufgap->gap_growth_size = element_count;
	return bufgap;
}


guint gdk_buffer_gap_trim_gap(GdkBufferGap *bufgap) {

	gulong new_byte_count;		/* num bytes to shrink buffer to */

	gconstpointer src;	/* Source of the memmove */
	gpointer dest;		/* Destination of the memmove */

	guint right_size;	/* Number of data elements in the right side */
	guint right_byte_count;	/* Number of bytes in the right side */


	g_return_val_if_fail(bufgap->gap_size > bufgap->gap_growth_size,
				bufgap->gap_size);

	right_size = gdk_buffer_gap_get_right_size(bufgap);

	/* Step 1: Move the right data left to the appropriate spot */
	/* Step 2: Shrink the buffer by the appropriate number of bytes */

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
	
		case gucharBuffer:
			/* Step 1: */
			right_byte_count = right_size * sizeof(guchar);
			src = (gconstpointer)(bufgap->buffer.ch 
						+ bufgap->gap_position
						+ bufgap->gap_size);
			dest = (gpointer)(bufgap->buffer.ch 
						+ bufgap->gap_position
						+ bufgap->gap_growth_size);
			g_memmove(dest, src, right_byte_count);
			/* Step 2: */
			new_byte_count = sizeof(guchar) * 
				(bufgap->gap_position + bufgap->gap_growth_size
				 + right_size);
			bufgap->buffer.ch = (guchar *)g_realloc(bufgap->buffer.wc, 
				new_byte_count);

			break;

		case GdkWCharBuffer:
			/* Step 1: */
			right_byte_count = right_size * sizeof(GdkWChar);
			src = (gconstpointer)(bufgap->buffer.wc 
						+ bufgap->gap_position
						+ bufgap->gap_size);
			dest = (gpointer)(bufgap->buffer.wc 
						+ bufgap->gap_position
						+ bufgap->gap_growth_size);
			g_memmove(dest, src, right_byte_count);
			/* Step 2: */
			new_byte_count = sizeof(GdkWChar) * 
				(bufgap->gap_position + bufgap->gap_growth_size
				 + right_size);
			bufgap->buffer.wc = (GdkWChar *)g_realloc(bufgap->buffer.wc, 
				new_byte_count);

			break;
			
		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return bufgap->gap_size;
	}
	/* End DATATYPE code */
	
	/* Step 3: Correct the applicable size member variables: */
	bufgap->buffer_size = (bufgap->gap_position + bufgap->gap_growth_size
				 + right_size);
	bufgap->gap_size = bufgap->gap_growth_size;
	
	return bufgap->gap_size;


}



/* Static function definitions: */
static GdkBufferGap *grow_gap(GdkBufferGap *bufgap, guint element_count) {

	gconstpointer src;	/* Source of the memmove */
	gpointer dest;		/* Destination of the memmove */

	guint right_size;	/* Number of data elements in the right side */
	guint right_byte_count;	/* Number of bytes in the right side */

	gulong bytes_to_add;		/* num bytes to grow buffer by */
	gulong orig_byte_count;		/* num bytes currently in buffer */
	

	right_size = gdk_buffer_gap_get_right_size(bufgap);

	/* Step 1: Grow the buffer by the appropriate number of bytes */
	/* Step 2: Move the right data half flush to the end of the buffer */

	/* DATATYPE code: */
	switch ( bufgap->buffer_type ) {
	
		case gucharBuffer:
			/* Step 1: */
			orig_byte_count = sizeof(guchar) * bufgap->buffer_size;
			bytes_to_add = sizeof(guchar) * element_count;
			bufgap->buffer.ch = (guchar *)g_realloc(bufgap->buffer.ch, 
				orig_byte_count + bytes_to_add);
			/* Step 2: */
			right_byte_count = right_size * sizeof(guchar);
			src = (gconstpointer)(bufgap->buffer.ch 
						+ bufgap->gap_position
						+ bufgap->gap_size);
			dest = (gpointer)(bufgap->buffer.ch 
						+ bufgap->gap_position
						+ bufgap->gap_size 
						+ element_count);
			g_memmove(dest, src, right_byte_count);

			break;

		case GdkWCharBuffer:
			/* Step 1: */
			orig_byte_count = sizeof(GdkWChar) * bufgap->buffer_size;
			bytes_to_add = sizeof(GdkWChar) * element_count;
			bufgap->buffer.wc = (GdkWChar *)g_realloc(bufgap->buffer.wc, 
				orig_byte_count + bytes_to_add);
			/* Step 2: */
			right_byte_count = right_size * sizeof(GdkWChar);
			src = (gconstpointer)(bufgap->buffer.wc 
						+ bufgap->gap_position
						+ bufgap->gap_size);
			dest = (gpointer)(bufgap->buffer.wc 
						+ bufgap->gap_position
						+ bufgap->gap_size 
						+ element_count);
			g_memmove(dest, src, right_byte_count);

			break;
			
		default:
			g_warning("Unrecognized GdkBufferGapBufferType! ");
			return NULL;
	}
	/* End DATATYPE code */
	
	/* Step 3: Correct the applicable size member variables: */
	bufgap->buffer_size += element_count;
	bufgap->gap_size += element_count;
		
	return bufgap;
}  /* End grow_gap */







