/*
Copyright (c) Derek Simkowiak 2000

GdkBufferGap

Author: Derek Simkowiak <dereks@kd-dev.com>

	This is an emacs-like gapped text buffer, as described at
http://www.web.us.uu.net/staff/djm/lore/buffer-gap .  The object is an
automatically-growing buffer with a gap in it.

	It is called GdkBufferGap because one of the character data types
it can hold is a GdkWChar (wide character), which requires Gdk, although
most people will probably want use just use the more common guchar
(unsigned char) data type.  I don't know why it's called a GdkWChar
instead of GWChar (?).  It should be very easy to add new character data
types, such as the upcoming GUChar in glib 1.4.

	This gapped buffer 'object' will automatically grow as necessary.
It also handles all the resizing of the gap, reallocating, copying/moving
of data, etc.  All the user needs to worry about is the position of the
gap (aka "insertion point").

	The buffer gives you separate access to the left and right sides of the
buffer, so if you want access to the entire buffer as a single array you can
either (a) move the gap to position zero and then read the right side of the
buffer, or (b) move the gap to the end of the buffer and then read the left side
of the buffer.  The reason for not doing this automatically for you is that
GNU regex operations are optimized for taking the left- and right- sides of a
gapped text buffer, and I wanted to keep the interface simple.
*/

#include <gdk/gdk.h>

typedef enum {

  gucharBuffer,		/* Holds guchar in the buffer */
  GdkWCharBuffer	/* Holds GdkWChar in the buffer */

} GdkBufferGapBufferType;

typedef struct _GdkBufferGap GdkBufferGap;

/* All members of this struct are private */
struct _GdkBufferGap {

  /* The buffer, either an array of GdkWChar or guchar */
  union { GdkWChar *wc; guchar  *ch; } buffer;

  /* The allocated length of the buffer, with gap (in elements, not  bytes). 
     buffer_size*sizeof(element type) is the number of bytes allocated.
     (buffer_size - gap_size) is the actual number of data elements. */
  guint buffer_size;

  /* Determines what kind of data this holds */
  GdkBufferGapBufferType buffer_type;

  /* The gap position, i.e. array index to insertion point */
  guint gap_position;

  /* The gap size, s.t. *(buffer + gap_position + gap_size) is
   * the first valid character following the gap.  Measured in
   * elements, not bytes */
  guint gap_size;

  /* The size of a new gap (because the gap got filled up) */
  guint gap_growth_size;
};

/* gdk_buffer_gap_new()
Description: Creates a new GdkBufferGap, with a default gap size.  A new default
gap size (used when the gap fills up) can be tweaked--see below.
Inputs: buffer_type: What kind of data you wish to hold in this buffer.
Outputs: A pointer to the new GdkBufferGap, or NULL if there's an error.
Preconditions: none.
Postconditions: The memory for the buffer gap is allocated (with default
values).  The gap is located at the beginning of the buffer.
*/
GdkBufferGap *gdk_buffer_gap_new(GdkBufferGapBufferType buffer_type);


/* gdk_buffer_gap_free()
Description: frees a buffer gap AND ITS DATA TOO.  If you have any
pointers to the data, and you call this function, you'll have dangling
ptrs.
Inputs: bufgap: The GdkBufferGap (and its data!) you want to free
Outputs: none.
Preconditions: bufgap should point to a GdkBufferGap.
Postconditions: The memory for the GdkBufferGap and its data is freed.
*/
void gdk_buffer_gap_free(GdkBufferGap *bufgap);


/* gdk_buffer_gap_insert()
Description: inserts the characters at the current insertion point.  Note that
if you insert a type of data other than what you specified to
gdk_buffer_gap_new(), you'll break things.  I.e., don't pass GdkWChar data to
a guchar GdkBufferGap.
Inputs: bufgap: The GdkBufferGap your inserting into.
	chars: The text data, NOT NULL-terminated, either an array of guchars or
		GdkWChars.
	num_chars: The number of characters you're inserting.
Outputs: The new GdkBufferGap, or NULL if there's an error.
Preconditions: bufgap should point to a valid GdkBufferGap.  chars should be
non-NULL.  num_chars should not exceed the memory space allocated for 'chars'.
Postconditions: The chars are copied into the buffer.  The gap position will
move to the right by num_chars.  The gap (and hence buffer) may also be resized,
but that's not for the user to worry about :)
*/
GdkBufferGap *gdk_buffer_gap_insert(GdkBufferGap *bufgap, 
					gconstpointer chars,
					guint num_chars );

/* gdk_buffer_gap_delete()
Description: Deletes text from the current insertion point.  It can delete
forward (i.e, enduser hits the DELETE key) or backward (i.e., enduser hits the
BACKSPACE key).  If you try to delete more characters than are physically
possible (like you delete 10 chars in a buffer that has only 5 chars, or you
delete 10 chars "forward" when the gap is as the very end of the buffer) then it
will silently delete as many characters as possible.
Inputs: bufgap: The GdkBufferGap you're deleting from
	num_chars: The number of characters to delete
	forward: If TRUE, deletes to the right.  If FALSE, deletes to the left.
Outputs: The number of characters actually deleted (will be <= num_chars).
Preconditions: bufgap should point to a valid GdkBufferGap.
Postconditions: The characters are "deleted" (and the gap grows by the returned
value.)
*/
guint gdk_buffer_gap_delete(GdkBufferGap *bufgap, 
				guint num_chars, 
				gboolean forward);

/* gdk_buffer_gap_move_gap()
Descriptions: Moves the gap to the specified position.  If the position is
greater than the size of the buffer, the position will simply be moved to the
end of the buffer.  Position zero "0" is the beginning of the buffer.
Inputs: bufgap: The GdkBufferGap whose gap you are moving
	position: The position you want to move the gap to.
Outputs: The actual position that the gap winds up at (will be <= position)
Preconditions: bufgap should point to a valid GdkBufferGap.
Postconditions: The insertion point (aka "gap") is at the returned position.
*/
guint gdk_buffer_gap_move_gap(GdkBufferGap *bufgap, guint position);


/* gdk_buffer_gap_get_left_pointer()
Description: Gets a pointer to the left side of the buffer.  Note that this will
be the address of element zero of the array, i.e., buffer[0].  You'll want to
use this in conjunction with gdk_buffer_gap_get_left_size() so that you don't
read the garbage in the gap.
Inputs: bufgap: The GdkBufferGap whose left-side pointer you want to get
Outputs: A read-only pointer to the left side of the buffer.
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: none.
*/
gconstpointer gdk_buffer_gap_get_left_pointer(GdkBufferGap *bufgap);

/* gdk_buffer_gap_get_left_size()
Description: Gets the size of the left side of the buffer.  You'll need this
when you use gdk_buffer_gap_get_left_pointer().  This value becomes invalid
as soon as you move the gap or change the buffer.
Inputs: bufgap: The GdkBufferGap whose left-side size you want to know.
Outputs: The number of elements (i.e. characters) between the beginning of the
buffer, and the position of the gap.  NOT necessarily equal to the number of 
bytes!
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: none.
*/
guint gdk_buffer_gap_get_left_size(GdkBufferGap *bufgap);


/* gdk_buffer_gap_get_right_pointer()
Description: Gets a pointer to the right side of the buffer.  Note that this 
will be the address directly after the end of the gap.  You'll want to
use this in conjunction with gdk_buffer_gap_get_right_size() so that you don't
read over the end of the array.  This value becomes invalidated as soon as you
change the buffer or gap.
Inputs: bufgap: The GdkBufferGap whose right-side pointer you want to get
Outputs: A read-only pointer to the right side of the buffer.
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: none.
*/
gconstpointer gdk_buffer_gap_get_right_pointer(GdkBufferGap *bufgap);


/* gdk_buffer_gap_get_right_size()
Description: Gets the size of the right side of the buffer.  You'll need this
when you use gdk_buffer_gap_get_right_pointer().  This value becomes invalid
as soon as you move the gap or change the buffer.
Inputs: bufgap: The GdkBufferGap whose right-side size you want to know.
Outputs: The number of elements (i.e. characters) between the end of the
gap, and the end of the buffer.  NOT necessarily equal to the number of 
bytes!
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: none.
*/
guint gdk_buffer_gap_get_right_size(GdkBufferGap *bufgap);


/* gdk_buffer_gap_get_data_size()
Description: Gives you the total number of valid data 
elements in the gapped text buffer.  This is not the same as the size of the
buffer itself, as it does not account for the gap in the buffer!
Inputs: bufgap: The GdkBufferGap whose total element count you want to know.
Outputs: The number of elements in the buffer.
Preconditions: bufgap should point to a valid GdkBufferGap.
Postconditions: none.
*/
#define gdk_buffer_gap_get_data_size(bufgap) \
	( (bufgap)->buffer_size - (bufgap)->gap_size )



/* Memory tuning */

/* gdk_buffer_gap_set_gap_growth_size()
Description: When you insert data into the buffer, the gap size will shrink 
by the size of your insertion.  When it gets to zero, a new gap is created 
and realloc is called.

	Of course, if your insertion is greater than the available gap size, the
GdkBufferGap will only grow the gap size once.  I.e., if the gap growth size is
1K, a one megabyte insertion will not result in over a thousand 1K-gap growths.

	So unless you are using the GdkBufferGap in an environment where the
average insertion is unusually large, you won't need to mess with the gap growth
size.  If you do need to set it to a particular size, this is the function to
use.
Inputs: bufgap: The GdkBufferGap whose gap growth size you want to set.
	element_count: The number of elements (NOT BYTES) that is a new gap's
	growth size.  Note that element_count*(sizeof(element type)) is the
	growth size in bytes.
Outputs: The new, tweaked GdkBufferGap
Preconditions: bufgap should point to a valid GdkBufferGap
Postconditions: the gap growth size is adjusted accordingly.  This will only
have an effect the next time the gap is re-grown.
*/
GdkBufferGap *gdk_buffer_gap_set_gap_growth_size(GdkBufferGap *bufgap, guint element_count);


/* gdk_buffer_gap_trim_gap()
Description: As you delete text from the buffer, the gap size will grow by the
amount you delete.  That's how a gapped text buffer works.

	But if you have a large buffer, and then delete most of the text,
your gap will very large in relation to your buffer.  Normally, this is fine
because the average size of a text file will stay the same during an editing
session.

	However, if you know your text buffer is going to remain small, you
would be wasting memory in the huge gap.  This function allows you to trim the
gap size down to a new gap's size (see above).  If the gap size is already
smaller than the gap growth size, nothing happens.
Inputs: bufgap: The GdkBufferGap whose gap you want to trim
Outputs: The size of the gap after any trimming is done.  Will be <= to the gap
growth size.
Preconditions: It only makes sense to call this function if your gap is huge in
relation to your buffer, and you don't expect the buffer to eventually get
filled up again, and you really need the memory.
Postconditions: The buffer's gap is shrunk down.
*/
guint gdk_buffer_gap_trim_gap(GdkBufferGap *bufgap);





